import re
import numpy as np
from glob import glob
from scipy.interpolate import interp2d, bisplev, interp1d, UnivariateSpline

def load_data(path):
    return np.loadtxt(path, dtype=np.double, delimiter=",", skiprows=5)
   

def get_data_map(data):
    x_labels, y_labels = get_xy(data)
    Nx = len(x_labels)
    Ny = len(y_labels)

    data_map = np.zeros([Ny,Nx])

    # Maybe it's better to be safe than smart?
    # indicies = np.zeros([np.shape(data)[0], 2], dtype=int)
    # indicies[:, 0] = data[:, 0] / np.min(x_labels) - 1
    # indicies[:, 1] = data[:, 1] / np.min(y_labels) - 1

    # Back to trying to be smart:
    indicies = np.array([np.repeat(np.arange(len(x_labels), dtype=int), len(y_labels)),
                         np.tile(np.arange(len(y_labels), dtype=int), len(x_labels))]).T


    for ij, val in zip(indicies, data[:, 2]):
        i, j = ij
        data_map[j, i] = val  # here we have a transposition
        # Why? Image is like: constant x = consant row, but first index changes row, so unintuitive.

    # Being smart = errros
    # for i in range(Ny):
    #     data_map[:, i] = data[Nx*i:Nx*(i+1), 2]
    
    return data_map

def get_xy(data):
    x = sorted(list(set(data[:, 0])))
    y = sorted(list(set(data[:, 1])))
    
    return x, y

def length_real2plot(labels, val):
    # Labels indicate the axis
    return val / (max(labels) - min(labels)) * len(labels)  # note that length will scale without shifting val

def coord_real2plot(xlabels, ylabels, xy):
    x, y = xy
    # Relative is in px (but can be float)
    # Here
    relative_x = (x - min(xlabels)) / (max(xlabels) - min(xlabels)) * len(xlabels) 
    relative_y = (y - min(ylabels)) / (max(ylabels) - min(ylabels)) * len(ylabels)
    # This weird substraction is related to the fact that each point corresponds to the middle of a pixel, so it shifts
    #   coordinate system by half of pixel.
    return np.array([relative_x, relative_y]) - np.array([0.5, 0.5])
    
def plot_geometry(xlabels, ylabels):
    # Assume correct axis labels
    d_scr = 200
    d_ch = 55
    # print(coord_real2plot(xlabels, ylabels, [0, d_scr]))
    # print(coord_real2plot(xlabels, ylabels, [0, d_scr+d_ch]))
    plt.axline(coord_real2plot(xlabels, ylabels, [min(xlabels), d_scr]), slope=0, color="black", ls="--")  # boundries of screening layer (not channel oxide layer!)
    plt.axline(coord_real2plot(xlabels, ylabels, [min(xlabels), d_scr+d_ch]), slope=0, color="black", ls="--")  # boundries of screening layer (not channel oxide layer!)

    w_block = 500
    w_met = 30
    h_met = 27
    h_ox = 10
    d_dot = 100
    # This weird substraction is related to the fact that each point corresponds to the middle of a pixel, so it shifts
    #   coordinate system by half of pixel.
    barrier = patches.Rectangle(coord_real2plot(xlabels, ylabels, [w_block/2-w_met/2, min(ylabels)]),
                                width=length_real2plot(xlabels, w_met),
                                height=length_real2plot(ylabels, d_scr+d_ch-h_ox+h_met-min(ylabels)),
                                fill=False,
                                ls="--",
                                color="black")
    plt.gca().add_patch(barrier)
    plungerL = patches.Rectangle(coord_real2plot(xlabels, ylabels, [w_block/2-w_met/2-d_dot/2, max(ylabels)]),
                                width=length_real2plot(xlabels, w_met),
                                height=-length_real2plot(ylabels, d_scr+d_ch-h_ox+h_met-min(ylabels)),
                                fill=False,
                                ls="--",
                                color="black")
    plt.gca().add_patch(plungerL)
    plungerR = patches.Rectangle(coord_real2plot(xlabels, ylabels, [w_block/2-w_met/2+d_dot/2, max(ylabels)]),
                                width=length_real2plot(xlabels, w_met),
                                height=-length_real2plot(ylabels, d_scr+d_ch-h_ox+h_met-min(ylabels)),
                                fill=False,
                                ls="--",
                                color="black")
    plt.gca().add_patch(plungerR)

def plot_map(ax, data_map, xlabels, ylabels, title, ticks_every=2, show_geometry=True,
             cbar=True, cbar_kwargs=[], vmin=None, vmax=None):
    plt.imshow(data_map, origin="lower", vmax=vmax, vmin=vmin)
    plt.xticks(np.arange(0, len(xlabels), ticks_every), labels=np.array(xlabels[::ticks_every]).astype(int))
    plt.yticks(np.arange(0, len(ylabels), ticks_every), labels=np.array(ylabels[::ticks_every]).astype(int))
    if show_geometry:
        plot_geometry(xlabels, ylabels)
        
    if cbar:
        plt.colorbar(fraction=0.046, pad=0.04, **cbar_kwargs)
    plt.xlabel("x (nm)")
    plt.ylabel("y (nm)")
    
    plt.title(title)

#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
def get_interpolated_map(data_map, xlabels, ylabels, dx=0, dy=0, Nx=200, Ny=200, return_func=False):
    interpolated_func = interp2d(xlabels, ylabels, data_map, kind="cubic")
    xspace = np.linspace(xlabels[0], xlabels[-1], Nx)
    yspace = np.linspace(ylabels[0], ylabels[-1], Ny)
    if not return_func:
        interpolated = interpolated_func(xspace, yspace, dx=dx, dy=dy)
        return xspace, yspace, interpolated
    else:
        interpolated_func_wrapper = lambda xspace, yspace: interpolated_func(xspace, yspace, dx=dx, dy=dy)
        return xspace, yspace, interpolated_func_wrapper
#!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
    
    
def get_z_from_path(path):
    return float(re.search("(?<=_z)([\d\.]*)(?=\.csv)", path)[0])  # debug here: regexr.com



class interpolate_maps():
    def __init__(self, data_maps, Nx, Ny, zlabels) -> None:
        self.Nx = Nx
        self.Ny = Ny
        self.interpolations = []
        for stack in data_maps:
            self.interpolations.append(UnivariateSpline(zlabels, stack))
        
    def __call__(self, z, dz=0):
        flat_map = []
        for func in self.interpolations:
            flat_map.append(func.derivative(n=dz)(z))

        return np.array(flat_map).reshape([self.Nx, self.Ny])

def interpolate_z(data_path):
    """ 
    Function for interpolating a whole data map in a z axis.
    Assumes standard naming convention (looks for _zXXXX.csv pattern).
    Returns interpolation functions that takes argument z and optional dz
        and returns data map.
    """
    
    paths = glob(data_path+"*.csv")
    print(paths)
    assert len(paths) > 3, "You need more than 3 z data points to make this interpolation."
    sort_idx = np.argsort([get_z_from_path(path) for path in paths])
    paths_sorted = np.array(paths)[sort_idx]

    # Get the dimensions assuming all are the same as in first data file
    data = load_data(paths[0])
    xlabels, ylabels = get_xy(data)
    Nx, Ny = (len(xlabels), len(ylabels))
    maps = np.zeros([Nx*Ny, len(paths_sorted)])

    # Read the data for interpolation
    zlabels = []
    for i, fname in enumerate(paths_sorted):
        data = load_data(fname)
        xlabels, ylabels = get_xy(data)
        data_map = get_data_map(data)
        z = get_z_from_path(fname)
        zlabels.append(z)
        # print(data_map.shape)
        maps[:, i] = data_map.flatten()

    return interpolate_maps(maps, Nx,Ny, zlabels)